import functools
import logging
import ssl
import sys
import time
from typing import Tuple, Optional

import anyio
import yaml
from anyio import create_tcp_listener, get_cancelled_exc_class, create_task_group
from anyio.streams.tls import TLSListener

from tiny_proxy import HttpProxyHandler, Socks4ProxyHandler, Socks5ProxyHandler

CLS_MAP = {
    'http': HttpProxyHandler,
    'socks4': Socks4ProxyHandler,
    'socks5': Socks5ProxyHandler,
}

logger = logging.getLogger(__name__)


def configure_logging():
    root_logger = logging.getLogger()
    root_logger.setLevel('INFO')

    fmt = '%(asctime)s [%(name)s:%(lineno)d] %(levelname)s : %(message)s'
    formatter = logging.Formatter(fmt)
    formatter.converter = time.gmtime

    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    root_logger.addHandler(stdout_handler)


def load_settings(file_name='./settings.yml') -> dict:
    with open(file_name, 'r') as f:
        return yaml.safe_load(f)


async def serve(
    proxy_type: str,
    host: str,
    port: int,
    ssl_cert: Optional[Tuple[str, str]] = None,
    **kwargs,
):
    handler_cls = CLS_MAP.get(proxy_type)
    if not handler_cls:
        raise RuntimeError(f'Unsupported proxy type: {proxy_type}')

    if ssl_cert is not None:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_context.load_cert_chain(*ssl_cert)
    else:
        ssl_context = None

    logger.info(f'Starting {proxy_type} proxy on {host}:{port}...')

    handler = handler_cls(**kwargs)

    try:
        listener = await create_tcp_listener(local_host=host, local_port=port)
        if ssl_context is not None:
            listener = TLSListener(listener=listener, ssl_context=ssl_context)

        async with listener:
            await listener.serve(handler.handle)
    except get_cancelled_exc_class():  # noqa
        pass


async def start_server(configs: list[dict]):
    async with create_task_group() as tg:
        for cfg in configs:
            tg.start_soon(functools.partial(serve, **cfg))


def main():
    configure_logging()
    settings = load_settings()
    try:
        anyio.run(start_server, settings['proxies'])
    except (KeyboardInterrupt, SystemExit):
        pass


if __name__ == '__main__':
    main()



import asyncio
import socket
import time
from typing import Iterable


def is_connectable(host, port):
    sock = None
    try:
        sock = socket.create_connection((host, port), timeout=1)
    except socket.error:
        return False
    else:
        return True
    finally:
        if sock is not None:
            sock.close()


def wait_until_connectable(host, port, timeout=10):
    count = 0
    while not is_connectable(host=host, port=port):
        if count >= timeout:
            raise Exception(
                f'The proxy server has not available by ({host}, {port}) in {timeout:d} seconds'
            )
        count += 1
        time.sleep(1)
    return True


def cancel_tasks(tasks: Iterable[asyncio.Task], loop: asyncio.AbstractEventLoop):
    if not tasks:
        return

    for task in tasks:
        task.cancel()

    loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))

    for task in tasks:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


def cancel_all_tasks(loop: asyncio.AbstractEventLoop):
    tasks = [task for task in asyncio.all_tasks(loop) if not task.done()]
    cancel_tasks(tasks=tasks, loop=loop)



import ssl

import pytest
import trustme

from tests.config import (
    TEST_HTTP_HOST_IPV4,
    TEST_HTTP_PORT_IPV4,
    PROXY_HOST,
    SOCKS5_PROXY_PORT,
    PROXY_USERNAME,
    PROXY_PASSWORD,
    SOCKS5_PROXY_PORT_NO_AUTH,
    SOCKS4_PROXY_PORT,
    SOCKS4_PROXY_PORT_NO_AUTH,
    HTTP_PROXY_PORT,
    HTTP_PROXY_PORT_NO_AUTH,
    TEST_HTTPS_HOST_IPV4,
    TEST_HTTPS_PORT_IPV4,
    TEST_HTTPS_HOST_IPV6,
    TEST_HTTPS_PORT_IPV6,
)
from tests.http_server import HttpServerConfig, HttpServer
from tests.proxy_server import ProxyConfig, ProxyServerRunner
from tests.utils import wait_until_connectable


@pytest.fixture(scope='session')
def ssl_ca() -> trustme.CA:
    return trustme.CA()


@pytest.fixture(scope='session')
def ssl_cert(ssl_ca: trustme.CA) -> trustme.LeafCert:
    return ssl_ca.issue_cert(
        "localhost",
        "127.0.0.1",
        "::1",
    )


@pytest.fixture(scope='session')
def ssl_certfile(ssl_cert: trustme.LeafCert):
    with ssl_cert.cert_chain_pems[0].tempfile() as cert_path:
        yield cert_path


@pytest.fixture(scope='session')
def ssl_keyfile(ssl_cert: trustme.LeafCert):
    with ssl_cert.private_key_pem.tempfile() as private_key_path:
        yield private_key_path


@pytest.fixture(scope='session')
def ssl_key_and_cert_chain_file(ssl_cert: trustme.LeafCert):
    with ssl_cert.private_key_and_cert_chain_pem.tempfile() as path:
        yield path


@pytest.fixture(scope='session')
def ssl_ca_cert_file(ssl_ca: trustme.CA):
    with ssl_ca.cert_pem.tempfile() as ca_cert_pem:
        yield ca_cert_pem


@pytest.fixture(scope='session')
def server_ssl_context(ssl_cert: trustme.LeafCert) -> ssl.SSLContext:
    ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    ssl_cert.configure_cert(ssl_ctx)
    return ssl_ctx


@pytest.fixture(scope='session')
def client_ssl_context(ssl_ca: trustme.CA, ssl_ca_cert_file) -> ssl.SSLContext:
    ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    ssl_ctx.verify_mode = ssl.CERT_REQUIRED
    ssl_ctx.check_hostname = True

    # ssl_ctx.load_verify_locations(ssl_ca_cert_file)
    ssl_ca.configure_trust(ssl_ctx)

    return ssl_ctx


@pytest.fixture(scope='session', autouse=True)
def web_server(ssl_certfile, ssl_keyfile):
    config = [
        HttpServerConfig(
            host=TEST_HTTP_HOST_IPV4,
            port=TEST_HTTP_PORT_IPV4,
        ),
        HttpServerConfig(
            host=TEST_HTTPS_HOST_IPV4,
            port=TEST_HTTPS_PORT_IPV4,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
        ),
        HttpServerConfig(
            host=TEST_HTTPS_HOST_IPV6,
            port=TEST_HTTPS_PORT_IPV6,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
        ),
    ]

    server = HttpServer(config=config)
    server.run()

    for cfg in config:
        wait_until_connectable(host=cfg.host, port=cfg.port)

    yield None

    server.shutdown()


@pytest.fixture(scope='session', autouse=True)
def proxy_server():
    config = [
        ProxyConfig(
            proxy_type='socks5',
            host=PROXY_HOST,
            port=SOCKS5_PROXY_PORT,
            username=PROXY_USERNAME,
            password=PROXY_PASSWORD,
        ),
        ProxyConfig(
            proxy_type='socks5',
            host=PROXY_HOST,
            port=SOCKS5_PROXY_PORT_NO_AUTH,
        ),
        ProxyConfig(
            proxy_type='socks4',
            host=PROXY_HOST,
            port=SOCKS4_PROXY_PORT,
            username=PROXY_USERNAME,
            password=None,
        ),
        ProxyConfig(
            proxy_type='socks4',
            host=PROXY_HOST,
            port=SOCKS4_PROXY_PORT_NO_AUTH,
        ),
        ProxyConfig(
            proxy_type='http',
            host=PROXY_HOST,
            port=HTTP_PROXY_PORT,
            username=PROXY_USERNAME,
            password=PROXY_PASSWORD,
        ),
        ProxyConfig(
            proxy_type='http',
            host=PROXY_HOST,
            port=HTTP_PROXY_PORT_NO_AUTH,
        ),
    ]

    server = ProxyServerRunner(config=config)
    server.run()
    for cfg in config:
        wait_until_connectable(host=cfg.host, port=cfg.port)

    yield None

    server.shutdown()



import logging

from .base import BaseProxyHandler
from .._proxy.abc import AbstractProxy
from .._proxy.socks5 import Socks5Proxy
from .._stream import SocketStream


class Socks5ProxyHandler(BaseProxyHandler):
    def __init__(
        self,
        username: str = None,
        password: str = None,
    ):
        self.username = username
        self.password = password
        self.logger = logging.getLogger(__name__)

    def create_proxy(self, stream: SocketStream) -> AbstractProxy:
        return Socks5Proxy(
            stream=stream,
            username=self.username,
            password=self.password,
        )



import enum
import logging
import ipaddress
from typing import Tuple

import anyio
import anyio.abc

from .._stream import SocketStream
from .._errors import ProxyError
from .abc import AbstractProxy

RSV = NULL = 0x00
SOCKS_VER4 = 0x04


class Command(enum.IntEnum):
    CONNECT = 0x01
    BIND = 0x02


class ReplyCode(enum.IntEnum):
    REQUEST_GRANTED = 0x5A
    REQUEST_REJECTED_OR_FAILED = 0x5B
    CONNECTION_FAILED = 0x5C
    AUTHENTICATION_FAILED = 0x5D


ReplyMessages = {
    ReplyCode.REQUEST_GRANTED: "Request granted",
    ReplyCode.REQUEST_REJECTED_OR_FAILED: "Request rejected or failed",
    ReplyCode.CONNECTION_FAILED: (
        "Request rejected because SOCKS server cannot connect to identd on the client"
    ),
    ReplyCode.AUTHENTICATION_FAILED: (
        "Request rejected because the client program and identd report different user-ids"
    ),
}


class Socks4Proxy(AbstractProxy):
    def __init__(self, stream: SocketStream, username: str = None):
        self.stream = stream
        self.username = username
        self.logger = logging.getLogger(__name__)

    async def connect_to_remote(self) -> SocketStream:
        try:
            remote_host, remote_port = await self.negotiate()
        except (
            anyio.EndOfStream,
            anyio.IncompleteRead,
            anyio.ClosedResourceError,
            anyio.BrokenResourceError,
        ) as e:
            raise ConnectionResetError(
                f'Connection reset by peer {self.stream.getpeername()}'
            ) from e

        local_addr = self.stream.getsockname()
        remote_addr = (remote_host, remote_port)
        self.logger.info('CONNECT {} -> {}'.format(local_addr, remote_addr))

        try:
            stream = await anyio.connect_tcp(remote_host=remote_host, remote_port=remote_port)
            remote = SocketStream(stream)
        except OSError as e:
            await self.respond(ReplyCode.CONNECTION_FAILED)
            raise ProxyError(f"Couldn't connect to host {remote_host}:{remote_port}") from e
        else:
            await self.respond(ReplyCode.REQUEST_GRANTED)
            return remote

    async def negotiate(self) -> Tuple[str, int]:
        version, command = await self.stream.receive_exactly(2)

        if version != SOCKS_VER4:
            await self.respond(ReplyCode.REQUEST_REJECTED_OR_FAILED)
            raise ProxyError("Invalid socks version")

        if command != Command.CONNECT:
            await self.respond(ReplyCode.REQUEST_REJECTED_OR_FAILED)
            raise ProxyError("Unsupported command")

        port = int.from_bytes(await self.stream.receive_exactly(2), "big")
        host_bytes = await self.stream.receive_exactly(4)

        include_hostname = host_bytes[:3] == bytes([NULL, NULL, NULL])

        user = (await self.read_until_null()).decode("ascii")
        if self.username and self.username != user:
            await self.respond(ReplyCode.AUTHENTICATION_FAILED)
            raise ProxyError("Authentication failed")

        if include_hostname:
            host = (await self.read_until_null()).decode("ascii")
        else:
            host = str(ipaddress.IPv4Address(host_bytes))

        return host, port

    async def read_until_null(self) -> bytes:
        data = bytearray()
        while True:
            byte = ord(await self.stream.receive_exactly(1))
            if byte == NULL:
                break
            data.append(byte)
        return data

    async def respond(self, code: ReplyCode):
        await self.stream.send(
            bytes(
                [
                    RSV,
                    code,
                    NULL,
                    NULL,
                    NULL,
                    NULL,
                    NULL,
                    NULL,
                ]
            )
        )



"""
python -m pytest tests --cov=./tiny_proxy --cov-report term-missing -s
"""
import ssl

import httpx
import pytest
from httpx import Response
from httpx_socks import AsyncProxyTransport

from tests.config import (
    SOCKS5_PROXY_URL,
    TEST_HTTPS_URL_IPV4,
    TEST_HTTPS_URL_IPV6,
    TEST_HTTP_URL_IPV4,
    SOCKS5_PROXY_URL_NO_AUTH,
    SOCKS4_PROXY_URL,
    SOCKS4_PROXY_URL_NO_AUTH,
    HTTP_PROXY_URL,
    HTTP_PROXY_URL_NO_AUTH,
)


async def fetch(
    proxy_url: str,
    target_url: str,
    proxy_ssl: ssl.SSLContext = None,
    target_ssl: ssl.SSLContext = None,
    timeout: httpx.Timeout = None,
    **kwargs,
) -> Response:

    transport = AsyncProxyTransport.from_url(
        proxy_url,
        proxy_ssl=proxy_ssl,
        verify=target_ssl,
        **kwargs,
    )
    async with httpx.AsyncClient(transport=transport) as client:
        res = await client.get(target_url, timeout=timeout)
        return res


@pytest.mark.parametrize('url', (TEST_HTTP_URL_IPV4,))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks5_proxy_http(url, rdns):
    res = await fetch(
        proxy_url=SOCKS5_PROXY_URL,
        target_url=url,
        rdns=rdns,
    )
    assert res.status_code == 200


@pytest.mark.parametrize('url', (TEST_HTTPS_URL_IPV4, TEST_HTTPS_URL_IPV6))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks5_proxy_https(client_ssl_context, url, rdns):
    res = await fetch(
        proxy_url=SOCKS5_PROXY_URL,
        target_url=url,
        target_ssl=client_ssl_context,
        rdns=rdns,
    )
    assert res.status_code == 200


@pytest.mark.parametrize('url', (TEST_HTTPS_URL_IPV4, TEST_HTTPS_URL_IPV6))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks5_proxy_https_no_auth(client_ssl_context, url, rdns):
    res = await fetch(
        proxy_url=SOCKS5_PROXY_URL_NO_AUTH,
        target_url=url,
        target_ssl=client_ssl_context,
        rdns=rdns,
    )
    assert res.status_code == 200


@pytest.mark.parametrize('url', (TEST_HTTPS_URL_IPV4,))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks4_proxy_https(client_ssl_context, url, rdns):
    res = await fetch(
        proxy_url=SOCKS4_PROXY_URL,
        target_url=url,
        target_ssl=client_ssl_context,
        rdns=rdns,
    )
    assert res.status_code == 200


@pytest.mark.parametrize('url', (TEST_HTTPS_URL_IPV4,))
@pytest.mark.parametrize('rdns', (True, False))
@pytest.mark.asyncio
async def test_socks4_proxy_https_no_auth(client_ssl_context, url, rdns):
    res = await fetch(
        proxy_url=SOCKS4_PROXY_URL_NO_AUTH,
        target_url=url,
        target_ssl=client_ssl_context,
        rdns=rdns,
    )
    assert res.status_code == 200


@pytest.mark.parametrize('url', (TEST_HTTPS_URL_IPV4,))
@pytest.mark.asyncio
async def test_http_proxy_https(client_ssl_context, url):
    res = await fetch(
        proxy_url=HTTP_PROXY_URL,
        target_url=url,
        target_ssl=client_ssl_context,
    )
    assert res.status_code == 200


@pytest.mark.parametrize('url', (TEST_HTTPS_URL_IPV4,))
@pytest.mark.asyncio
async def test_http_proxy_https_no_auth(client_ssl_context, url):
    res = await fetch(
        proxy_url=HTTP_PROXY_URL_NO_AUTH,
        target_url=url,
        target_ssl=client_ssl_context,
    )
    assert res.status_code == 200



from ._errors import ProxyError
from ._stream import SocketStream
from ._tunnel import create_tunnel

from ._proxy.abc import AbstractProxy
from ._proxy.socks5 import Socks5Proxy
from ._proxy.socks4 import Socks4Proxy
from ._proxy.http import HttpProxy

from ._handlers.http import HttpProxyHandler
from ._handlers.socks4 import Socks4ProxyHandler
from ._handlers.socks5 import Socks5ProxyHandler

__version__ = '0.2.1'

__all__ = (
    'ProxyError',
    'SocketStream',
    'create_tunnel',
    'AbstractProxy',
    'Socks5Proxy',
    'Socks4Proxy',
    'HttpProxy',
    'HttpProxyHandler',
    'Socks4ProxyHandler',
    'Socks5ProxyHandler',
)



TEST_HTTP_HOST_IPV4 = '127.0.0.1'
TEST_HTTP_PORT_IPV4 = 8881
TEST_HTTP_URL_IPV4 = f'http://{TEST_HTTP_HOST_IPV4}:{TEST_HTTP_PORT_IPV4}/'

TEST_HTTPS_HOST_IPV4 = '127.0.0.1'
TEST_HTTPS_PORT_IPV4 = 8882
TEST_HTTPS_URL_IPV4 = f'https://{TEST_HTTPS_HOST_IPV4}:{TEST_HTTPS_PORT_IPV4}/'

TEST_HTTPS_HOST_IPV6 = '::1'
TEST_HTTPS_PORT_IPV6 = 8883
TEST_HTTPS_URL_IPV6 = f'https://[{TEST_HTTPS_HOST_IPV6}]:{TEST_HTTPS_PORT_IPV6}/'

PROXY_USERNAME = 'username'
PROXY_PASSWORD = 'password'

PROXY_HOST = '127.0.0.1'

SOCKS5_PROXY_PORT = 7780
SOCKS5_PROXY_PORT_NO_AUTH = 7781

SOCKS4_PROXY_PORT = 7782
SOCKS4_PROXY_PORT_NO_AUTH = 7783

HTTP_PROXY_PORT = 7784
HTTP_PROXY_PORT_NO_AUTH = 7785

SOCKS5_PROXY_URL = 'socks5://{username}:{password}@{host}:{port}'.format(
    host=PROXY_HOST,
    port=SOCKS5_PROXY_PORT,
    username=PROXY_USERNAME,
    password=PROXY_PASSWORD,
)

SOCKS5_PROXY_URL_NO_AUTH = 'socks5://{host}:{port}'.format(
    host=PROXY_HOST,
    port=SOCKS5_PROXY_PORT_NO_AUTH,
)

SOCKS4_PROXY_URL = 'socks4://{username}:{password}@{host}:{port}'.format(
    host=PROXY_HOST,
    port=SOCKS4_PROXY_PORT,
    username=PROXY_USERNAME,
    password='',
)

SOCKS4_PROXY_URL_NO_AUTH = 'socks4://{host}:{port}'.format(
    host=PROXY_HOST,
    port=SOCKS4_PROXY_PORT_NO_AUTH,
)

HTTP_PROXY_URL = 'http://{username}:{password}@{host}:{port}'.format(
    host=PROXY_HOST,
    port=HTTP_PROXY_PORT,
    username=PROXY_USERNAME,
    password=PROXY_PASSWORD,
)

HTTP_PROXY_URL_NO_AUTH = 'http://{host}:{port}'.format(
    host=PROXY_HOST,
    port=HTTP_PROXY_PORT_NO_AUTH,
)



import ssl

from aiohttp import web


async def index(_):
    return web.Response(body='Index')


app = web.Application()
app.router.add_get('/', index)


def run_app(
    host: str,
    port: int,
    ssl_certfile: str = None,
    ssl_keyfile: str = None,
):
    if ssl_certfile and ssl_keyfile:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        ssl_context.load_cert_chain(ssl_certfile, ssl_keyfile)
    else:
        ssl_context = None

    web.run_app(
        app,
        host=host,
        port=port,
        ssl_context=ssl_context,
        print=False,  # type: ignore
    )



import logging

from .base import BaseProxyHandler
from .._proxy.abc import AbstractProxy
from .._proxy.socks4 import Socks4Proxy
from .._stream import SocketStream


class Socks4ProxyHandler(BaseProxyHandler):
    def __init__(self, username: str = None):
        self.username = username
        self.logger = logging.getLogger(__name__)

    def create_proxy(self, stream: SocketStream) -> AbstractProxy:
        return Socks4Proxy(stream=stream, username=self.username)



import asyncio
import logging
import ssl
import typing
from contextlib import contextmanager
from multiprocessing import Process

from anyio import create_tcp_listener
from anyio.streams.tls import TLSListener

from tests.utils import cancel_all_tasks, cancel_tasks, wait_until_connectable
from tiny_proxy import HttpProxyHandler, Socks5ProxyHandler, Socks4ProxyHandler


class ProxyConfig(typing.NamedTuple):
    proxy_type: str
    host: str
    port: int
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None
    ssl_certfile: typing.Optional[str] = None
    ssl_keyfile: typing.Optional[str] = None

    def to_dict(self):
        d = {}
        for key, val in self._asdict().items():
            if val is not None:
                d[key] = val
        return d


class ProxyServer:
    cls_map = {
        'http': HttpProxyHandler,
        'socks4': Socks4ProxyHandler,
        'socks5': Socks5ProxyHandler,
    }

    def __init__(self, config: typing.Iterable[ProxyConfig], loop: asyncio.AbstractEventLoop):
        self.loop = loop
        self.config = config
        self.logger = logging.getLogger(__name__)
        self.server_tasks = []

    def run(self):
        proxies = self.config
        for proxy in proxies:
            server_task = self.loop.create_task(self._listen(**proxy.to_dict()))
            self.server_tasks.append(server_task)

    def run_forever(self):
        self.run()
        self.loop.run_forever()

    def shutdown(self):
        print('Shutting down...')

        cancel_tasks(self.server_tasks, self.loop)
        cancel_all_tasks(self.loop)

        self.loop.run_until_complete(self.loop.shutdown_asyncgens())
        try:
            self.loop.run_until_complete(self.loop.shutdown_default_executor())
        except AttributeError:  # pragma: no cover
            pass  # shutdown_default_executor is new to Python 3.9

        self.loop.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.shutdown()

    async def _listen(
        self,
        proxy_type,
        host,
        port,
        ssl_certfile=None,
        ssl_keyfile=None,
        **kwargs,
    ):
        handler_cls = self.cls_map.get(proxy_type)
        if not handler_cls:
            raise RuntimeError(f'Unsupported type: {proxy_type}')

        if ssl_certfile and ssl_keyfile:
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            ssl_context.load_cert_chain(ssl_certfile, ssl_keyfile)
        else:
            ssl_context = None

        print(f'Starting {proxy_type} proxy on {host}:{port}...')

        handler = handler_cls(**kwargs)

        listener = await create_tcp_listener(local_host=host, local_port=port)
        if ssl_context is not None:
            listener = TLSListener(listener=listener, ssl_context=ssl_context)

        async with listener:
            await listener.serve(handler.handle)


def _start_proxy_server(config: typing.Iterable[ProxyConfig]):
    import asyncio

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    server = ProxyServer(config=config, loop=loop)

    try:
        server.run_forever()
    except (KeyboardInterrupt, SystemExit):
        pass
    finally:
        server.shutdown()


class ProxyServerRunner:
    def __init__(self, config: typing.Iterable[ProxyConfig]):
        self.config = config
        self.process = None

    def run(self):
        """
        https://pytest-cov.readthedocs.io/en/latest/subprocess-support.html#if-you-use-multiprocessing-process
        or use Thread
        """
        try:
            from pytest_cov.embed import cleanup_on_sigterm  # noqa
        except ImportError:
            pass
        else:
            cleanup_on_sigterm()

        self.process = Process(target=_start_proxy_server, kwargs=dict(config=self.config))
        self.process.daemon = True
        self.process.start()

    def shutdown(self):
        self.process.terminate()


@contextmanager
def start_proxy_server(config: ProxyConfig):
    """
    https://pytest-cov.readthedocs.io/en/latest/subprocess-support.html#if-you-use-multiprocessing-process
    or use Thread
    """
    try:
        from pytest_cov.embed import cleanup_on_sigterm  # noqa
    except ImportError:
        pass
    else:
        cleanup_on_sigterm()

    process = Process(target=_start_proxy_server, kwargs=dict(config=[config]))
    # process = Thread(target=_start_proxy_server, kwargs=dict(config=[config]))
    process.daemon = True
    process.start()

    wait_until_connectable(host=config.host, port=config.port)

    try:
        yield None
    finally:
        process.terminate()
        process.join()
        pass



import enum
import ipaddress
import logging

import anyio
import anyio.abc

from .._stream import SocketStream
from .._errors import ProxyError
from .abc import AbstractProxy

RSV = NULL = 0x00
SOCKS_VER5 = 0x05

SOCKS5_GRANTED = 0x00


class AuthMethod(enum.IntEnum):
    ANONYMOUS = 0x00
    GSSAPI = 0x01
    USERNAME_PASSWORD = 0x02
    NO_ACCEPTABLE = 0xFF


class AddressType(enum.IntEnum):
    IPV4 = 0x01
    DOMAIN = 0x03
    IPV6 = 0x04

    @classmethod
    def from_ip_ver(cls, ver: int):
        if ver == 4:
            return cls.IPV4
        if ver == 6:
            return cls.IPV6

        raise ValueError('Invalid IP version')


class Command(enum.IntEnum):
    CONNECT = 0x01
    BIND = 0x02
    UDP_ASSOCIATE = 0x03


class ReplyCode(enum.IntEnum):
    SUCCEEDED = 0x00
    GENERAL_FAILURE = 0x01
    CONNECTION_NOT_ALLOWED = 0x02
    NETWORK_UNREACHABLE = 0x03
    HOST_UNREACHABLE = 0x04
    CONNECTION_REFUSED = 0x05
    TTL_EXPIRED = 0x06
    COMMAND_NOT_SUPPORTED = 0x07
    ADDRESS_TYPE_NOT_SUPPORTED = 0x08


ReplyMessages = {
    ReplyCode.SUCCEEDED: 'Request granted',
    ReplyCode.GENERAL_FAILURE: 'General SOCKS server failure',
    ReplyCode.CONNECTION_NOT_ALLOWED: 'Connection not allowed by ruleset',
    ReplyCode.NETWORK_UNREACHABLE: 'Network unreachable',
    ReplyCode.HOST_UNREACHABLE: 'Host unreachable',
    ReplyCode.CONNECTION_REFUSED: 'Connection refused by destination host',
    ReplyCode.TTL_EXPIRED: 'TTL expired',
    ReplyCode.COMMAND_NOT_SUPPORTED: 'Command not supported or protocol error',
    ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED: 'Address type not supported',
}


class Socks5Proxy(AbstractProxy):
    def __init__(self, stream: SocketStream, username=None, password=None):
        self.stream = stream
        self.username = username
        self.password = password
        self.logger = logging.getLogger(__name__)

    async def connect_to_remote(self) -> SocketStream:
        try:
            remote_host, remote_port = await self.negotiate()
        except (
            anyio.EndOfStream,
            anyio.IncompleteRead,
            anyio.ClosedResourceError,
            anyio.BrokenResourceError,
        ) as e:
            raise ConnectionResetError(
                f'Connection reset by peer {self.stream.getpeername()}'
            ) from e

        local_addr = self.stream.getsockname()
        remote_addr = (remote_host, remote_port)
        self.logger.info('CONNECT {} -> {}'.format(local_addr, remote_addr))

        try:
            # todo: add timeout?
            stream = await anyio.connect_tcp(remote_host=remote_host, remote_port=remote_port)
            remote = SocketStream(stream)
        except OSError as e:
            reply = bytes([SOCKS_VER5, ReplyCode.CONNECTION_REFUSED, NULL, NULL, NULL, NULL])
            await self.stream.send(reply)
            raise ProxyError(f"Couldn't connect to host {remote_host}:{remote_port}") from e
        else:
            bind_address = remote.getsockname()
            bind_ip = ipaddress.ip_address(bind_address[0])
            bind_port = bind_address[1]

            reply = bytearray(
                [
                    SOCKS_VER5,
                    ReplyCode.SUCCEEDED,
                    RSV,
                    AddressType.from_ip_ver(bind_ip.version),
                ]
            )
            reply += bind_ip.packed
            reply += bind_port.to_bytes(2, 'big')

            await self.stream.send(reply)

            return remote

    async def negotiate(self):
        auth_required = self.username and self.password

        # ------------------------AUTH METHODS---------------------
        version, num_methods = await self.stream.receive_exactly(2)

        if version != SOCKS_VER5:
            await self.stream.send(bytes([NULL, NULL]))
            raise ProxyError('Unsupported socks version')

        methods = []
        for i in range(num_methods):
            methods.append(ord(await self.stream.receive_exactly(1)))

        if auth_required:
            auth_method = (
                AuthMethod.USERNAME_PASSWORD
                if AuthMethod.USERNAME_PASSWORD in methods
                else AuthMethod.NO_ACCEPTABLE
            )
        else:
            auth_method = (
                AuthMethod.ANONYMOUS
                if AuthMethod.ANONYMOUS in methods
                else AuthMethod.NO_ACCEPTABLE
            )

        await self.stream.send(bytes([SOCKS_VER5, auth_method]))
        if auth_method == AuthMethod.NO_ACCEPTABLE:
            raise ProxyError('Not acceptable auth method')

        # -----------------------AUTH REQUEST-----------------------
        if auth_method == AuthMethod.USERNAME_PASSWORD:
            version = ord(await self.stream.receive_exactly(1))
            if version != 1:
                await self.stream.send(bytes([version, 0xFF]))
                raise ProxyError('Invalid auth request')

            username_len = ord(await self.stream.receive_exactly(1))
            username = (await self.stream.receive_exactly(username_len)).decode('utf-8')

            password_len = ord(await self.stream.receive_exactly(1))
            password = (await self.stream.receive_exactly(password_len)).decode('utf-8')

            if username == self.username and password == self.password:
                await self.stream.send(bytes([version, SOCKS5_GRANTED]))
            else:
                await self.stream.send(bytes([version, 0xFF]))
                raise ProxyError('Authentication failed')

        # --------------------------CONNECT-----------------------------
        version, cmd, _, address_type = await self.stream.receive_exactly(4)

        if version != SOCKS_VER5:
            await self.stream.send(bytes([SOCKS_VER5, ReplyCode.GENERAL_FAILURE, RSV]))
            raise ProxyError(ReplyMessages[ReplyCode.GENERAL_FAILURE])

        if cmd != Command.CONNECT:
            await self.stream.send(bytes([SOCKS_VER5, ReplyCode.COMMAND_NOT_SUPPORTED, RSV]))
            raise ProxyError(ReplyMessages[ReplyCode.COMMAND_NOT_SUPPORTED])

        if address_type == AddressType.IPV4:
            remote_host = str(ipaddress.IPv4Address(await self.stream.receive_exactly(4)))
        elif address_type == AddressType.IPV6:
            remote_host = str(ipaddress.IPv6Address(await self.stream.receive_exactly(16)))
        elif address_type == AddressType.DOMAIN:
            domain_length = ord(await self.stream.receive_exactly(1))
            remote_host = (await self.stream.receive_exactly(domain_length)).decode('ascii')
        else:
            await self.stream.send(
                bytes([SOCKS_VER5, ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED, NULL, NULL, NULL, NULL])
            )
            raise ProxyError(ReplyMessages[ReplyCode.ADDRESS_TYPE_NOT_SUPPORTED])

        remote_port = int.from_bytes(await self.stream.receive_exactly(2), 'big')

        return remote_host, remote_port



import typing
from multiprocessing import Process

from tests.http_app import run_app


class HttpServerConfig(typing.NamedTuple):
    host: str
    port: int
    ssl_certfile: str = None
    ssl_keyfile: str = None

    def to_dict(self):
        d = {}
        for key, val in self._asdict().items():
            if val is not None:
                d[key] = val
        return d


class HttpServer:
    def __init__(self, config: typing.Iterable[HttpServerConfig]):
        self.config = config
        self.workers = []

    def run(self):
        for cfg in self.config:
            print(f'Starting web server on {cfg.host}:{cfg.port}...')
            p = Process(target=run_app, kwargs=cfg.to_dict())
            self.workers.append(p)

        for p in self.workers:
            p.start()

    def shutdown(self):
        for p in self.workers:
            p.terminate()



import anyio
import anyio.abc
from anyio.streams.buffered import BufferedByteReceiveStream

DEFAULT_RECEIVE_SIZE = 65536


class SocketStream:
    def __init__(self, stream: anyio.abc.SocketStream):
        self._stream = stream
        self._buffered = BufferedByteReceiveStream(stream)
        self._closing = False

    async def send(self, data: bytes) -> None:
        await self._stream.send(data)

    async def send_eof(self) -> None:
        await self._stream.send_eof()

    async def receive(self, max_bytes=DEFAULT_RECEIVE_SIZE) -> bytes:
        return await self._buffered.receive(max_bytes)

    async def receive_exactly(self, n) -> bytes:
        return await self._buffered.receive_exactly(n)

    async def receive_until(self, delimiter: bytes, max_bytes: int) -> bytes:
        return await self._buffered.receive_until(delimiter, max_bytes)

    async def aclose(self):
        if not self._closing:
            self._closing = True
            try:
                # underlying TLSStream.aclose() -> TLSStream.unwrap()
                await self._buffered.aclose()
            except (anyio.BrokenResourceError, anyio.BusyResourceError):
                pass

    def getpeername(self):
        return self._stream.extra(anyio.abc.SocketAttribute.remote_address, '')

    def getsockname(self):
        return self._stream.extra(anyio.abc.SocketAttribute.local_address, '')



import anyio

from ._stream import SocketStream, DEFAULT_RECEIVE_SIZE


async def create_tunnel(endpoint1: SocketStream, endpoint2: SocketStream):
    async def pipe(reader: SocketStream, writer: SocketStream):
        try:
            while True:
                try:
                    data = await reader.receive(DEFAULT_RECEIVE_SIZE)
                except (
                    anyio.EndOfStream,
                    anyio.ClosedResourceError,
                    anyio.BrokenResourceError,
                ):
                    break

                try:
                    await writer.send(data)
                except (
                    anyio.ClosedResourceError,
                    anyio.BrokenResourceError,
                ):
                    break
        finally:
            await writer.aclose()

    async with anyio.create_task_group() as tg:
        tg.start_soon(pipe, endpoint1, endpoint2)
        tg.start_soon(pipe, endpoint2, endpoint1)



import logging
from typing import Union

import anyio
import anyio.abc
from anyio.streams.tls import TLSStream

from .._stream import SocketStream
from .._proxy.abc import AbstractProxy
from .._tunnel import create_tunnel

AnyioSocketStream = Union[anyio.abc.SocketStream, TLSStream]


class BaseProxyHandler:
    logger: logging.Logger

    async def handle(self, stream: AnyioSocketStream):
        client = SocketStream(stream)
        proxy = self.create_proxy(client)

        try:
            remote = await proxy.connect_to_remote()
        except anyio.get_cancelled_exc_class():  # noqa
            await client.aclose()
        except Exception as e:
            await client.aclose()
            self.logger.error(e)
            self.logger.debug(e, exc_info=True)
        else:
            try:
                await create_tunnel(client, remote)
            except anyio.get_cancelled_exc_class():  # noqa  # pragma: nocover
                pass
            except Exception as e:  # pragma: nocover
                self.logger.error(e)
                self.logger.debug(e, exc_info=True)
            finally:
                await remote.aclose()
                await client.aclose()

    def create_proxy(self, stream: SocketStream) -> AbstractProxy:
        raise NotImplementedError()



import base64
import binascii
import logging
from collections import namedtuple
from http.server import BaseHTTPRequestHandler
from io import BytesIO
from typing import Tuple

import anyio
import anyio.abc

from .abc import AbstractProxy
from .._errors import ProxyError
from .._stream import SocketStream


class HTTPRequest(BaseHTTPRequestHandler):
    """
    https://stackoverflow.com/questions/4685217/parse-raw-http-headers
    """

    # noinspection PyMissingConstructor
    def __init__(self, data: bytes):
        self.rfile = BytesIO(data)
        self.raw_requestline = self.rfile.readline()
        self.error_code = self.error_message = None
        self.parse_request()

    def send_error(self, code, message=None, explain=None):
        self.error_code = code


class BasicAuth(namedtuple('BasicAuth', ['login', 'password', 'encoding'])):
    """Http basic authentication helper."""

    def __new__(cls, login: str, password: str = '', encoding: str = 'latin1') -> 'BasicAuth':
        if login is None:
            raise ValueError('None is not allowed as login value')

        if password is None:
            raise ValueError('None is not allowed as password value')

        if ':' in login:
            raise ValueError('A ":" is not allowed in login (RFC 1945#section-11.1)')

        # noinspection PyTypeChecker,PyArgumentList
        return super().__new__(cls, login, password, encoding)

    @classmethod
    def decode(cls, auth_header: str, encoding: str = 'latin1') -> 'BasicAuth':
        """Create a BasicAuth object from an Authorization HTTP header."""
        try:
            auth_type, encoded_credentials = auth_header.split(' ', 1)
        except ValueError:
            raise ValueError('Could not parse authorization header.')

        if auth_type.lower() != 'basic':
            raise ValueError('Unknown authorization method %s' % auth_type)

        try:
            decoded = base64.b64decode(encoded_credentials.encode('ascii'), validate=True).decode(
                encoding
            )
        except binascii.Error:
            raise ValueError('Invalid base64 encoding.')

        try:
            # RFC 2617 HTTP Authentication
            # https://www.ietf.org/rfc/rfc2617.txt
            # the colon must be present, but the username and password may be
            # otherwise blank.
            username, password = decoded.split(':', 1)
        except ValueError:
            raise ValueError('Invalid credentials.')

        # noinspection PyTypeChecker
        return cls(username, password, encoding=encoding)

    def encode(self) -> str:
        """Encode credentials."""
        creds = ('%s:%s' % (self.login, self.password)).encode(self.encoding)
        return 'Basic %s' % base64.b64encode(creds).decode(self.encoding)


class HttpProxy(AbstractProxy):
    def __init__(
        self,
        stream: SocketStream,
        username: str = None,
        password: str = None,
    ):
        self.stream = stream
        self.username = username
        self.password = password
        self.logger = logging.getLogger(__name__)

    async def connect_to_remote(self) -> SocketStream:
        try:
            remote_host, remote_port = await self.negotiate()
        except (
            anyio.EndOfStream,
            anyio.IncompleteRead,
            anyio.ClosedResourceError,
            anyio.BrokenResourceError,
        ) as e:
            raise ConnectionResetError(
                f'Connection reset by peer {self.stream.getpeername()}'
            ) from e

        local_addr = self.stream.getsockname()
        remote_addr = (remote_host, remote_port)
        self.logger.info('CONNECT {} -> {}'.format(local_addr, remote_addr))

        try:
            stream = await anyio.connect_tcp(remote_host=remote_host, remote_port=remote_port)
            remote = SocketStream(stream)
        except OSError as e:
            self.logger.error(e)
            await self.respond(502, 'Bad Gateway', raise_exc=False)
            raise ProxyError(f"Couldn't connect to host {remote_host}:{remote_port}") from e
        else:
            await self.respond(200, 'Connection established')
            return remote

    async def negotiate(self) -> Tuple[str, int]:
        data = await self.stream.receive_until(b'\r\n\r\n', 4096)
        req = HTTPRequest(data)

        if req.error_code:
            await self.respond(int(req.error_code), req.error_message)

        if req.command is None or req.command.lower() != 'connect':
            self.logger.debug(repr(data))
            await self.respond(400, 'Bad Request')

        if self.username and self.password:
            auth_header = req.headers['proxy-authorization']

            if not auth_header:
                await self.respond(401, 'Unauthorized')

            try:
                auth = BasicAuth.decode(auth_header)
            except ValueError:
                await self.respond(401, 'Unauthorized')
            else:
                if auth.login != self.username or auth.password != self.password:
                    await self.respond(401, 'Unauthorized')

        try:
            host, port = req.path.split(":")
            port = int(port)
        except ValueError:
            await self.respond(400, 'Bad Request')
            raise

        return host, port

    async def respond(self, code: int, message: str, raise_exc=True):
        res = f'HTTP/1.1 {code} {message}\r\n\r\n'
        await self.stream.send(res.encode('ascii'))
        if code != 200 and raise_exc:
            raise ProxyError(f'{code} {message}')



#!/usr/bin/env python
import os
import re
import sys

from setuptools import setup


def get_version():
    here = os.path.dirname(os.path.abspath(__file__))
    filename = os.path.join(here, 'tiny_proxy', '__init__.py')
    contents = open(filename).read()
    pattern = r"^__version__ = '(.*?)'$"
    return re.search(pattern, contents, re.MULTILINE).group(1)


def get_long_description():
    with open('README.md', mode='r', encoding='utf8') as f:
        return f.read()


if sys.version_info < (3, 7):
    raise RuntimeError('tiny-proxy requires Python 3.7+')


setup(
    name='tiny_proxy',
    author='Roman Snegirev',
    author_email='snegiryev@gmail.com',
    version=get_version(),
    license='Apache 2',
    url='https://github.com/romis2012/tiny-proxy',
    description='Simple proxy server (SOCKS4(a), SOCKS5(h), HTTP tunnel)',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=[
        'tiny_proxy',
        'tiny_proxy._proxy',
        'tiny_proxy._handlers',
    ],
    keywords='socks socks5 socks4 http proxy server asyncio trio anyio',
    install_requires=[
        'anyio>=3.6.1,<5.0',
    ],
)
